import { startService } from 'esbuild';
import * as colors from 'kleur/colors';
import path from 'path';
import { promises as fs } from 'fs';
let esbuildService = null;
const IS_PREACT = /from\s+['"]preact['"]/;
function checkIsPreact(filePath, contents) {
    return filePath.endsWith('.jsx') && IS_PREACT.test(contents);
}
function getLoader(filePath) {
    const ext = path.extname(filePath);
    if (ext === '.mjs') {
        return 'js';
    }
    return ext.substr(1);
}
export function esbuildPlugin(_, { input }) {
    return {
        name: '@snowpack/plugin-esbuild',
        resolve: {
            input,
            output: ['.js'],
        },
        async load({ filePath }) {
            esbuildService = esbuildService || (await startService());
            const contents = await fs.readFile(filePath, 'utf-8');
            const isPreact = checkIsPreact(filePath, contents);
            const { js, warnings } = await esbuildService.transform(contents, {
                loader: getLoader(filePath),
                jsxFactory: isPreact ? 'h' : undefined,
                jsxFragment: isPreact ? 'Fragment' : undefined,
            });
            for (const warning of warnings) {
                console.error(colors.bold('! ') + filePath);
                console.error('  ' + warning.text);
            }
            return { '.js': js || '' };
        },
    };
}
export function stopEsbuild() {
    esbuildService && esbuildService.stop();
}
